// Copyright Marcus Del Favero 2025
// Licensed under the GNU AGPLv3 with an exception, see `README.md` for details
#[cfg(feature = "client")] use glam::Vec2;

use super::{DMap, DSet, Player, PlayerId, bullet::Bullet, powerup::{PowerupType, Effect}, event::move_chooser::MoveType, team::TeamId};
#[cfg(feature = "client")] use super::player::{PlayerColour, direction::Direction};

#[cfg(feature = "client")] use crate::utils::maths::Circle;

/*
 * Both the client and server need to know certain things that have changed
 * after updating the world. However, the exact information that's relevant to
 * them differs.
 *
 * As an optimisation, generics are used to avoid performing unnecessary
 * operations when calling `update`.
 */
pub trait Changes {
	/*
	 * For the client, prevents changes reported to old players from being
	 * visualised on players (in case if a player rapidly exits and joins).
	 *
	 * For the server, probably not needed for now but I'm adding it for good
	 * practice to clean things up when players are removed.
	 */
	fn removed_players(&mut self, old_ids: DSet<PlayerId>);

	fn add_hit(&mut self, bullet: Bullet, player: &Player, player_id: PlayerId, killed: bool);
	fn pos_changed(&mut self, id: PlayerId);
	fn ammo_crate_collected(&mut self, id: PlayerId);
	fn powerup_collected(&mut self, id: PlayerId, player: &Player, typ: PowerupType);
	fn player_reset(&mut self, id: PlayerId);
	fn player_team_change(&mut self, player_id: PlayerId, team_id: Option<TeamId>);
	fn items_reset(&mut self);
}

#[cfg(feature = "client")]
pub struct Hit {
	pub bullet: Bullet,
	pub player_id: Option<PlayerId>,
	pub player_pos: Vec2,
	pub player_vel: Vec2,
	pub player_dir: Direction,
	pub player_colour: PlayerColour,
	pub killed: bool,
}

#[cfg(feature = "client")]
pub struct ClientChanges { // Used by the clients
	current_id: PlayerId,
	pub moved: DSet<PlayerId>, // Whenever the server moves a player, which can be on death or teleportation
	pub hits: Vec<Hit>,
	pub reset: DSet<PlayerId>,
	pub items_reset: bool,
	pub current_team_change: Option<TeamId>,
}

#[cfg(feature = "client")]
impl ClientChanges {
	pub fn new(current_id: PlayerId) -> ClientChanges {
		ClientChanges {
			current_id,
			moved: DSet::new(),
			hits: Vec::new(),
			reset: DSet::new(),
			items_reset: false,
			current_team_change: None,
		}
	}
}

#[cfg(feature = "client")]
impl Changes for ClientChanges {
	fn removed_players(&mut self, old_ids: DSet<PlayerId>) {
		for id in &old_ids {
			self.moved.remove(id);
			self.reset.remove(id);
		}

		for hit in &mut self.hits {
			if let Some(id) = hit.player_id {
				if old_ids.contains(&id) {
					hit.player_id = None;
				}
			}
		}
	}

	fn add_hit(&mut self, bullet: Bullet, player: &Player, player_id: PlayerId, killed: bool) {
		self.hits.push(Hit {
			bullet,
			player_id: Some(player_id),
			player_pos: player.get_pos(),
			player_vel: player.get_vel(),
			player_dir: player.get_dir(),
			player_colour: player.get_colour(),
			killed,
		});
	}

	fn pos_changed(&mut self, id: PlayerId) { self.moved.insert(id); }
	fn ammo_crate_collected(&mut self, _id: PlayerId) {}
	fn powerup_collected(&mut self, _id: PlayerId, _player: &Player, _typ: PowerupType) {}
	fn player_reset(&mut self, id: PlayerId) { self.reset.insert(id); }
	fn items_reset(&mut self) { self.items_reset = true; }

	fn player_team_change(&mut self, player_id: PlayerId, team_id: Option<TeamId>) {
		if player_id == self.current_id {
			self.current_team_change = team_id.or(self.current_team_change);
		}
	}
}

#[derive(Default)]
pub struct ServerChanges {
	pub kills: Vec<Kill>, // Player ids duplicated if they get multiple kills in a single update
	pub(super) randomly_moved: DMap<PlayerId, (MoveType, Option<TeamId>)>,
	pub(super) dead: DSet<PlayerId>,
	pub(super) ammo_crates_collected: Vec<PlayerId>,
	pub(super) effects_to_apply: Vec<(PlayerId, Effect)>,
}

pub struct Kill {
	pub killer: Option<PlayerId>,
	pub killed: PlayerId,
}

impl Changes for ServerChanges {
	fn removed_players(&mut self, old_ids: DSet<PlayerId>) {
		for id in &old_ids {
			self.randomly_moved.remove(id);
			self.dead.remove(id);
		}

		self.ammo_crates_collected.retain(|id| !old_ids.contains(id));
		self.effects_to_apply.retain(|(id, _)| !old_ids.contains(id));
		// Not changing self.kills as that is used before the players are eliminated
	}

	fn add_hit(&mut self, bullet: Bullet, player: &Player, player_id: PlayerId, killed: bool) {
		if killed {
			self.kills.push(Kill {
				killer: bullet.get_player_owner(),
				killed: player_id,
			});

			self.randomly_moved.insert(player_id, (MoveType::Spawn, player.get_team()));
			self.dead.insert(player_id);
		}
	}

	fn pos_changed(&mut self, _id: PlayerId) {}
	fn ammo_crate_collected(&mut self, id: PlayerId) { self.ammo_crates_collected.push(id); }

	fn powerup_collected(&mut self, id: PlayerId, player: &Player, typ: PowerupType) {
		if let Some(effect) = Effect::try_from_type(typ) {
			self.effects_to_apply.push((id, effect));
		} else if typ == PowerupType::Teleportation {
			self.randomly_moved.insert(id, (MoveType::Teleportation, player.get_team()));
		}
	}

	fn player_reset(&mut self, _id: PlayerId) {}
	fn items_reset(&mut self) {}
	fn player_team_change(&mut self, _player_id: PlayerId, _team_id: Option<TeamId>) {}
}

pub struct IgnoreChanges;

impl Changes for IgnoreChanges {
	fn removed_players(&mut self, _old_ids: DSet<PlayerId>) {}
	fn add_hit(&mut self, _bullet: Bullet, _player: &Player, _player_id: PlayerId, _killed: bool) {}
	fn pos_changed(&mut self, _id: PlayerId) {}
	fn ammo_crate_collected(&mut self, _id: PlayerId) {}
	fn powerup_collected(&mut self, _id: PlayerId, _player: &Player, _typ: PowerupType) {}
	fn player_reset(&mut self, _id: PlayerId) {}
	fn items_reset(&mut self) {}
	fn player_team_change(&mut self, _player_id: PlayerId, _team_id: Option<TeamId>) {}
}
