// Copyright Marcus Del Favero 2025
// Licensed under the GNU AGPLv3 with an exception, see `README.md` for details
mod local_player;

pub use local_player::LocalPlayer;

use std::{num::NonZeroUsize, collections::{VecDeque, btree_map::Entry}};

use glam::Vec2;

use super::{
	World, PlayerId, TeamId, DMap, DELTA_TIME, MAX_STEPS, changes::ClientChanges,
	player::{Player, DPlayer, update::{PlayerUpdate, PlayerBulkUpdate},
	forcefield::Forcefield}, bullet::{Bullet, BulletManager}, powerup::Effect,
};

use crate::net::{client::ClientError, message::Message};
use crate::playing::{model::Model, view::{Camera, HealthBar, HealthBarRenderer}};
use crate::blocks::Blocks;
use crate::utils::{RollingAverage, maths::decay::Decay};
use crate::game::config::action::{Action, ActionEvent, ActionState};
use crate::playing::view::NameAbovePlayer;

pub struct SmoothWorld {
	current_id: PlayerId,

	/**
	 * Initialised to None and is set to Some(team_id) when the current player
	 * joins a team.
	 *
	 * This is used to remember the team the player was in while spectating so
	 * that team's scores can be indented.
	 */
	last_current_team: Option<TeamId>,
	followed_id: Option<PlayerId>, // INVARIANT: If followed_id == Some(x), then x ∈ world
	prev_followed_id: PlayerId, // Used when restoring to a followed player
	players: DMap<PlayerId, LocalPlayer>,
	bullets: BulletManager,

	/**
	 * Updates that have been performed locally that the server hasn't yet
	 * acknowledged.
	 *
	 * Storing this allows the player to receive immediate feedback on moving
	 * around even if there's large ping.
	 */
	staging_updates: VecDeque<PlayerUpdate>,
	past_staging_updates: RollingAverage,

	/**
	 * The number of updates that have been sent out. This is the sequence number
	 * of the update at the back of the staging update queue.
	 */
	update_count: u64,

	/**
	 * When the client is too far ahead or behind the server, don't immediately
	 * jump the smooth world in time as that looks bad. Instead, keep the error
	 * but exponentially decay it towards zero so things are smoother.
	 */
	time_shift: f32,
}

impl SmoothWorld {
	pub fn new() -> SmoothWorld {
		SmoothWorld {
			current_id: PlayerId::default(),
			last_current_team: None,
			followed_id: None,
			prev_followed_id: PlayerId::default(),
			players: DMap::new(),
			bullets: BulletManager::new(),
			staging_updates: VecDeque::new(),
			past_staging_updates: RollingAverage::new(const { NonZeroUsize::new(256).unwrap() }),
			update_count: 0,
			time_shift: 0.0,
		}
	}

	pub fn reset(&mut self, model: &Model) {
		self.current_id = model.get_id();
		self.last_current_team = model.get_world().get_player_team(self.current_id);
		self.followed_id = model.get_world().has_player(self.current_id).then_some(self.current_id);
		self.prev_followed_id = PlayerId::default();
		self.players = model.get_world().players.iter().map(|(&id, player)| (id, LocalPlayer::new(player.clone()))).collect();
		self.bullets = model.get_world().bullets.clone();
		self.staging_updates.clear();
		self.past_staging_updates.reset();
		self.update_count = 0;
		self.time_shift = 0.0;
	}

	pub fn player_count(&self) -> usize { self.players.len() }

	pub fn get_total_time_shift(&self) -> f32 {
		self.staging_updates.len() as f32 * DELTA_TIME + self.time_shift
	}

	pub fn player_iter(&self) -> impl Iterator<Item = &LocalPlayer> {
		self.players.values()
	}

	pub(super) fn player_iter_mut(&mut self) -> impl Iterator<Item = (PlayerId, &mut LocalPlayer)> {
		self.players.iter_mut().map(|(&id, player)| (id, player))
	}

	pub fn followed_player(&self) -> Option<(PlayerId, &LocalPlayer)> {
		self.followed_id.map(|id| (id, &self.players[&id]))
	}

	pub fn followed_team(&self) -> Option<TeamId> {
		self.followed_id.map_or(self.last_current_team, |id| self.players[&id].inner().get_team())
	}

	pub fn current_player_mut(&mut self) -> Option<&mut LocalPlayer> {
		self.players.get_mut(&self.current_id)
	}

	pub(super) fn get_bullets(&self) -> &[Bullet] { &self.bullets }

	pub fn get_forcefields(&self, world: &World, time_diff: f32) -> Vec<(Forcefield, f32)> {
		let time_shift = self.get_total_time_shift() + time_diff;

		/*
		 * Subtracting time_diff as this time decreases over time rather than
		 * increases (it's the remaining time of the effect).
		 */
		world.effects
			.player_iter_with_effect(Effect::Forcefield)
			.map(|(player_id, info)| {
				let player = &self.players[&player_id];
				(Forcefield { player_id, team_id: player.inner().get_team(), pos: player.get_pos(time_diff), strength: info.power }, info.time - time_shift)
			})
			.collect()
	}

	pub fn create_bulk_update(&self) -> Message {
		// update_count is the sequence number of the end update
		// Instead want that for the beginning update
		let sequence_number = self.oldest_sequence_number();
		let mut bulk_update = PlayerBulkUpdate::new(sequence_number);
		for &update in &self.staging_updates {
			bulk_update.push(update);
		}
		Message::PlayerBulkUpdate(bulk_update)
	}

	fn oldest_sequence_number(&self) -> u64 {
		self.update_count - (self.staging_updates.len() - 1) as u64
	}

	/**
	 * Returns the next valid player id that is greater than or equal to the
	 * input id, wrapping around if nothing's found or returning None if there
	 * are no ids.
	 */
	fn valid_id_at_or_after(&self, id: PlayerId) -> Option<PlayerId> {
		self.players.range(id..).next().map(|(&id, _)| id).or_else(|| self.players.first_key_value().map(|e| *e.0))
	}

	/**
	 * Performs an action related to changing the followed player while
	 * spectating. Returns the position of the newly followed player, if any.
	 */
	pub fn action(&mut self, action: ActionEvent) -> Option<Vec2> {
		// Only when spectating
		if self.players.contains_key(&self.current_id) { return None; }

		let prev_followed = self.followed_id;

		match action {
			ActionEvent::Action(Action::FollowPlayer, ActionState::Pressed) => {
				self.followed_id = if self.followed_id.is_some() {
					None
				} else {
					self.valid_id_at_or_after(self.prev_followed_id)
				};
			},
			ActionEvent::Action(Action::PrevPlayer, ActionState::Pressed) => {
				if let Some(id) = self.followed_id {
					self.followed_id = self.players.range(..id).next_back().map(|(&id, _)| id).or_else(|| self.players.last_key_value().map(|e| *e.0));
				}
			},
			ActionEvent::Action(Action::NextPlayer, ActionState::Pressed) => {
				if let Some(id) = self.followed_id {
					self.followed_id = self.valid_id_at_or_after(id.next());
				}
			},
			ActionEvent::Action(..) | ActionEvent::AllReleased => (),
		}

		if let Some(id) = self.followed_id {
			self.prev_followed_id = id;
		}

		self.followed_player().and_then(|(id, player)| (prev_followed != Some(id)).then_some(player.get_pos(0.0 /* Not perfect but I'm lazy */)))
	}

	pub fn remove_old_players(&mut self, world: &World) {
		self.players.retain(|id, _| world.players.contains_key(id));
	}

	/**
	 * Returns whether the camera position should be updated if the followed id
	 * changes.
	 */
	#[must_use]
	pub fn update_followed_player(&mut self) -> bool {
		if self.players.contains_key(&self.current_id) {
			if self.followed_id != Some(self.current_id) {
				self.followed_id = Some(self.current_id); // Follows the current player when they join
				return true; // Only moves the camera if this has changed
			}
		} else if let Some(id) = self.followed_id {
			if id == self.current_id {
				self.followed_id = None; // When a player enters spectating mode, don't have them follow any player
			} else if !self.players.contains_key(&id) {
				self.followed_id = self.valid_id_at_or_after(id.next()); // Satisfies the invariant if the player suddenly leaves
				if let Some(id) = self.followed_id {
					self.prev_followed_id = id;
				}
				return true;
			}
		}

		false
	}

	pub fn sync(&mut self, world: &World, blocks: &Blocks, changes: &ClientChanges, ack: u64, recent_update_count: usize) {
		let to_keep = self.update_count.checked_sub(ack).unwrap_or_else(|| {
			log::warn!("server acknowledged update {ack} which hasn't yet been sent (sequence number of most recent update = {})", self.update_count);
			0
		});

		let to_remove = self.staging_updates.len().checked_sub(to_keep as usize).unwrap_or_else(|| {
			log::warn!(
				"server acknowledged update {ack} which is earlier than the oldest staging update ({}), staging update count = {}, most recent update's sequence number = {}",
				self.oldest_sequence_number(),
				self.staging_updates.len(),
				self.update_count
			);

			0
		});

		self.time_shift += (to_remove as f32 - recent_update_count as f32) * DELTA_TIME;

		for _ in 0..to_remove { self.staging_updates.pop_front(); }

		self.bullets = world.bullets.clone();

		/*
		 * Extrapolates the current world's state by the length of the staging
		 * update queue.
		 *
		 * For the current player this is applying the staging updates, and for
		 * all other players this is using the most recent update, as an
		 * estimation that likely will be the next one.
		 *
		 * Also updating the bullets into the future while doing this.
		 */
		let mut new_players = world.players.clone();
		for &update in &self.staging_updates {
			for (&id, player) in &mut new_players {
				if id == self.current_id { // Use the most recent update (which was already set) as a prediction if not the current player
					player.set_update(update);
				}

				SmoothWorld::update_player(&mut self.bullets, world, blocks, player, id);
			}

			SmoothWorld::update_bullets(&mut self.bullets, DELTA_TIME, world, blocks, new_players.iter_mut());
		}
		SmoothWorld::update_bullets(&mut self.bullets, self.time_shift, world, blocks, new_players.iter_mut());

		for (&id, player) in &new_players {
			match self.players.entry(id) {
				Entry::Vacant(e) => _ = e.insert(LocalPlayer::new(player.clone())),
				Entry::Occupied(mut e) => {
					let prev_player = e.get_mut();

					if !changes.moved.contains(&id) {
						let delta = &prev_player.player - player;
						prev_player.delta += &delta;
						if id == self.current_id {
							prev_player.delta.reset_angle();
						}
					} else {
						prev_player.delta = DPlayer::zero();
					}

					prev_player.player = player.clone();
				}
			}
		}

		for hit in &changes.hits {
			let Some(id) = hit.player_id else { continue; };
			if let Some(player) = self.players.get_mut(&id) { // Shouldn't fail but being safe
				player.last_hurt = if hit.killed { f32::INFINITY } else { 0.0 };
				if hit.killed {
					player.hbar.reset();
				}
			}
		}

		for id in &changes.reset {
			if let Some(player) = self.players.get_mut(id) {
				player.hbar.reset();
			}
		}

		self.last_current_team = changes.current_team_change.or(self.last_current_team);
	}

	pub fn update(&mut self, update: PlayerUpdate, world: &World, blocks: &Blocks) -> Result<(), ClientError> {
		if let Some(player) = self.players.get_mut(&self.current_id) {
			player.player.set_update(update);
		}

		self.staging_updates.push_back(update);
		self.update_count += 1;

		if self.staging_updates.len() >= MAX_STEPS {
			return Err(ClientError::Limit(String::from("too many staging updates accumulated by client")));
		}

		self.past_staging_updates.add(self.staging_updates.len() as f32);
		let mean_len = self.past_staging_updates.mean().unwrap();
		let decay_rate = (100.0 / (mean_len + 1.0)) + 1.0;
		let prev_time_shift = self.time_shift;
		self.time_shift.decay(2.0, DELTA_TIME);

		for (&id, player) in &mut self.players {
			player.delta.decay(decay_rate);
			SmoothWorld::update_player(&mut self.bullets, world, blocks, &mut player.player, id);
		}

		/*
		 * NOTE: In the future I can investigate adding the time shift to the
		 * players similar to what I do with the bullets. If done properly it
		 * could make things a bit smoother on bad network conditions.
		 */
		SmoothWorld::update_bullets(&mut self.bullets, DELTA_TIME - prev_time_shift + self.time_shift, world, blocks, self.players.iter_mut().map(|(id, player)| (id, &mut player.player)));

		Ok(())
	}

	fn update_player(bullets: &mut BulletManager, world: &World, blocks: &Blocks, player: &mut Player, id: PlayerId) {
		player.update(DELTA_TIME, world.config.size, blocks, &world.effects, id);
		player.decrease_cooldown(DELTA_TIME * world.effects.get_mul(id, Effect::Reload));
		if player.shoot() {
			bullets.add(Bullet::new(player, id, world.effects.get_mul(id, Effect::Damage), 0.0));
		}
	}

	fn update_bullets<'a>(bullets: &mut BulletManager, dt: f32, world: &World, blocks: &Blocks, players: impl Iterator<Item = (&'a PlayerId, &'a mut Player)>) {
		bullets.update(dt, &world.get_forcefields());
		for (&id, player) in players {
			for bullet in bullets.collide_player(player, id, world.config.friendly_fire) {
				player.on_bullet_hit(&bullet);
			}
		}
		bullets.collide_blocks(blocks);
	}

	/**
	 * Returns the list of information about how to render the player names.
	 */
	pub fn push_health_bars(&self, hbar_renderer: &mut HealthBarRenderer, lwinsize: Vec2, camera: &Camera, time_diff: f32) -> Vec<NameAbovePlayer> {
		// Render this player's health bar on top of all others
		let mut names = Vec::with_capacity(self.players.len());
		let builder = HealthBar::above_player_builder(lwinsize, camera);
		for (&id, player) in &self.players {
			if Some(id) != self.followed_id {
				let (hbar, name_pos) = builder.make(player, &player.hbar, time_diff);
				hbar_renderer.push(hbar);
				names.push(NameAbovePlayer::new(name_pos, player.player.get_style()));
			}
		}

		if let Some(id) = self.followed_id {
			hbar_renderer.push(self.players[&id].hbar.make_hud(lwinsize));
		}

		names
	}
}
